@@ -15,13 +15,13 @@ mod versions;
 
 use crate::bitcoincore_rpc::jsonrpc::serde_json::Value;
 use bitcoincore_rpc::{Auth, Client, RpcApi};
-use log::debug;
+use log::{debug, error, warn};
 use std::ffi::OsStr;
 use std::net::{Ipv4Addr, SocketAddrV4, TcpListener};
 use std::path::PathBuf;
 use std::process::{Child, Command, ExitStatus, Stdio};
 use std::time::Duration;
-use std::{env, fmt, thread};
+use std::{env, fmt, fs, thread};
 use tempfile::TempDir;
 
 pub extern crate bitcoincore_rpc;
@@ -33,14 +33,32 @@ pub struct BitcoinD {
     process: Child,
     /// Rpc client linked to this bitcoind process
     pub client: Client,
-    /// Work directory, where the node store blocks and other stuff. It is kept in the struct so that
-    /// directory is deleted only when this struct is dropped
-    _work_dir: TempDir,
+    /// Work directory, where the node store blocks and other stuff.
+    work_dir: DataDir,
 
     /// Contains information to connect to this node
     pub params: ConnectParams,
 }
 
+/// The DataDir struct defining the kind of data directory the node
+/// will contain. Data directory can be either persistent, or temporary.
+pub enum DataDir {
+    /// Persistent Data Directory
+    Persistent(PathBuf),
+    /// Temporary Data Directory
+    Temporary(TempDir),
+}
+
+impl DataDir {
+    /// Return the data directory path
+    fn path(&self) -> PathBuf {
+        match self {
+            Self::Persistent(path) => path.to_owned(),
+            Self::Temporary(tmp_dir) => tmp_dir.path().to_path_buf(),
+        }
+    }
+}
+
 #[derive(Debug, Clone)]
 /// Contains all the information to connect to this node
 pub struct ConnectParams {
@@ -55,7 +73,7 @@ pub struct ConnectParams {
 }
 
 /// Enum to specify p2p settings
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Debug, PartialEq, Eq, Clone)]
 pub enum P2P {
     /// the node doesn't open a p2p port and work in standalone mode
     No,
@@ -81,6 +99,10 @@ pub enum Error {
     NeitherFeatureNorEnvVar,
     /// Returned when calling methods requiring either a feature or anv var, but both are present
     BothFeatureAndEnvVar,
+    /// Wrapper of early exit status
+    EarlyExit(ExitStatus),
+    /// Returned when both tmpdir and staticdir is specified in `Conf` options
+    BothDirsSpecified,
 }
 
 impl fmt::Debug for Error {
@@ -92,6 +114,8 @@ impl fmt::Debug for Error {
             Error::NoEnvVar => write!(f, "Called a method requiring env var `BITCOIND_EXE` to be set, but it's not"),
             Error::NeitherFeatureNorEnvVar =>  write!(f, "Called a method requiring env var `BITCOIND_EXE` or a feature to be set, but neither are set"),
             Error::BothFeatureAndEnvVar => write!(f, "Called a method requiring env var `BITCOIND_EXE` or a feature to be set, but both are set"),
+            Error::EarlyExit(e) => write!(f, "The bitcoind process terminated early with exit code {}", e),
+            Error::BothDirsSpecified => write!(f, "tempdir and staticdir cannot be enabled at same time in configuration options")
         }
     }
 }
@@ -120,11 +144,13 @@ const LOCAL_IP: Ipv4Addr = Ipv4Addr::new(127, 0, 0, 1);
 /// conf.p2p = bitcoind::P2P::No;
 /// conf.network = "regtest";
 /// conf.tmpdir = None;
+/// conf.staticdir = None;
+/// conf.attempts = 3;
 /// assert_eq!(conf, bitcoind::Conf::default());
 /// ```
 ///
 #[non_exhaustive]
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Debug, PartialEq, Eq, Clone)]
 pub struct Conf<'a> {
     /// Bitcoind command line arguments containing no spaces like `vec!["-dbcache=300", "-regtest"]`
     /// note that `port`, `rpcport`, `connect`, `datadir`, `listen`
@@ -141,12 +167,30 @@ pub struct Conf<'a> {
     /// directory with different/esoteric networks
     pub network: &'a str,
 
-    /// Optionally specify the root of where the temporary directories will be created.
-    /// If none and the env var `TEMPDIR_ROOT` is set, the env var is used.
-    /// If none and the env var `TEMPDIR_ROOT` is not set, the default temp dir of the OS is used.
-    /// It may be useful for example to set to a ramdisk so that bitcoin nodes spawn very fast
-    /// because their datadirs are in RAM
+    /// Optionally specify a temporary or persistent working directory for the node.
+    /// The following two parameters can be configured to simulate desired working directory configuration.
+    ///
+    /// tmpdir is Some() && staticdir is Some() : Error. Cannot be enabled at same time.
+    /// tmpdir is Some(temp_path) && staticdir is None : Create temporary directory at `tmpdir` path.
+    /// tmpdir is None && staticdir is Some(work_path) : Create persistent directory at `staticdir` path.
+    /// tmpdir is None && staticdir is None: Creates a temporary directory in OS default temporary directory (eg /tmp) or `TEMPDIR_ROOT` env variable path.
+    ///
+    /// It may be useful for example to set to a ramdisk via `TEMPDIR_ROOT` env option so that
+    /// bitcoin nodes spawn very fast because their datadirs are in RAM. Should not be enabled with persistent
+    /// mode, as it cause memory overflows.
+
+    /// Temporary directory path
     pub tmpdir: Option<PathBuf>,
+
+    /// Persistent directory path
+    pub staticdir: Option<PathBuf>,
+
+    /// Try to spawn the process `attempt` time
+    ///
+    /// The OS is giving available ports to use, however, they aren't booked, so it could rarely
+    /// happen they are used at the time the process is spawn. When retrying other available ports
+    /// are returned reducing the probability of conflicts to negligible.
+    pub attempts: u8,
 }
 
 impl Default for Conf<'_> {
@@ -157,6 +201,8 @@ impl Default for Conf<'_> {
             p2p: P2P::No,
             network: "regtest",
             tmpdir: None,
+            staticdir: None,
+            attempts: 3,
         }
     }
 }
@@ -171,16 +217,19 @@ impl BitcoinD {
 
     /// Launch the bitcoind process from the given `exe` executable with given [Conf] param
     pub fn with_conf<S: AsRef<OsStr>>(exe: S, conf: &Conf) -> Result<BitcoinD, Error> {
-        let work_dir = match &conf.tmpdir {
-            Some(path) => TempDir::new_in(path),
-            None => match env::var("TEMPDIR_ROOT") {
-                Ok(env_path) => TempDir::new_in(env_path),
-                Err(_) => TempDir::new(),
-            },
-        }?;
-        debug!("work_dir: {:?}", work_dir);
-        let datadir = work_dir.path().to_path_buf();
-        let cookie_file = datadir.join(conf.network).join(".cookie");
+        let work_dir = match (&conf.tmpdir, &conf.staticdir) {
+            (Some(_), Some(_)) => return Err(Error::BothDirsSpecified),
+            (Some(tmpdir), None) => DataDir::Temporary(TempDir::new_in(tmpdir)?),
+            (None, Some(workdir)) => {
+                fs::create_dir_all(workdir)?;
+                DataDir::Persistent(workdir.to_owned())
+            }
+            (None, None) => DataDir::Temporary(TempDir::new()?),
+        };
+
+        let work_dir_path = work_dir.path();
+        debug!("work_dir: {:?}", work_dir_path);
+        let cookie_file = work_dir_path.join(conf.network).join(".cookie");
         let rpc_port = get_available_port()?;
         let rpc_socket = SocketAddrV4::new(LOCAL_IP, rpc_port);
         let rpc_url = format!("http://{}", rpc_socket);
@@ -211,7 +260,7 @@ impl BitcoinD {
             Stdio::null()
         };
 
-        let datadir_arg = format!("-datadir={}", datadir.display());
+        let datadir_arg = format!("-datadir={}", work_dir_path.display());
         let rpc_arg = format!("-rpcport={}", rpc_port);
         let default_args = [&datadir_arg, &rpc_arg];
 
@@ -221,7 +270,7 @@ impl BitcoinD {
             default_args,
             p2p_args
         );
-        let process = Command::new(exe)
+        let mut process = Command::new(exe.as_ref())
             .args(&default_args)
             .args(&p2p_args)
             .args(&conf.args)
@@ -231,6 +280,17 @@ impl BitcoinD {
         let node_url_default = format!("{}/wallet/default", rpc_url);
         // wait bitcoind is ready, use default wallet
         let client = loop {
+            if let Some(status) = process.try_wait()? {
+                if conf.attempts > 0 {
+                    warn!("early exit with: {:?}. Trying to launch again ({} attempts remaining), maybe some other process used our available port", status, conf.attempts);
+                    let mut conf = conf.clone();
+                    conf.attempts -= 1;
+                    return Self::with_conf(exe, &conf);
+                } else {
+                    error!("early exit with: {:?}", status);
+                    return Err(Error::EarlyExit(status));
+                }
+            }
             thread::sleep(Duration::from_millis(500));
             assert!(process.stderr.is_none());
             let client_result = Client::new(&rpc_url, Auth::CookieFile(cookie_file.clone()));
@@ -251,9 +311,9 @@ impl BitcoinD {
         Ok(BitcoinD {
             process,
             client,
-            _work_dir: work_dir,
+            work_dir,
             params: ConnectParams {
-                datadir,
+                datadir: work_dir_path,
                 cookie_file,
                 rpc_socket,
                 p2p_socket,
@@ -277,6 +337,11 @@ impl BitcoinD {
         )
     }
 
+    /// Return the current workdir path of the running node
+    pub fn workdir(&self) -> PathBuf {
+        self.work_dir.path()
+    }
+
     /// Returns the [P2P] enum to connect to this node p2p port
     pub fn p2p_connect(&self, listen: bool) -> Option<P2P> {
         self.params.p2p_socket.map(|s| P2P::Connect(s, listen))
@@ -304,6 +369,9 @@ impl BitcoinD {
 
 impl Drop for BitcoinD {
     fn drop(&mut self) {
+        if let DataDir::Persistent(_) = self.work_dir {
+            let _ = self.client.stop();
+        }
         let _ = self.process.kill();
     }
 }
@@ -373,7 +441,6 @@ mod test {
     #[test]
     fn test_bitcoind() {
         let exe = init();
-        println!("{}", exe);
         let bitcoind = BitcoinD::new(exe).unwrap();
         let info = bitcoind.client.get_blockchain_info().unwrap();
         assert_eq!(0, info.blocks);
@@ -418,6 +485,7 @@ mod test {
 
     #[test]
     fn test_multi_p2p() {
+        let _ = env_logger::try_init();
         let mut conf_node1 = Conf::default();
         conf_node1.p2p = P2P::Yes;
         let node1 = BitcoinD::with_conf(exe_path().unwrap(), &conf_node1).unwrap();
